@@ -14,14 +14,14 @@
 #    limitations under the License.
 
 from oslo_utils import uuidutils
-import six
+from osprofiler import profiler
 
 from mistral import exceptions as exc
 from mistral.lang import types
 from mistral.lang.v2 import base
 from mistral.lang.v2 import task_defaults
 from mistral.lang.v2 import tasks
-from mistral import utils
+from mistral_lib import utils
 
 NOOP_COMMAND = 'noop'
 FAIL_COMMAND = 'fail'
@@ -74,7 +74,7 @@ class WorkflowSpec(base.BaseSpec):
 
         # Inject 'type' here, so instantiate_spec function can recognize the
         # specific subclass of TaskSpec.
-        for task in six.itervalues(self._data.get('tasks')):
+        for task in self._data.get('tasks').values():
             task['type'] = self._type
 
         self._tasks = self._spec_property('tasks', tasks.TaskSpecList)
@@ -157,7 +157,7 @@ class DirectWorkflowSpec(WorkflowSpec):
                 "type": "object",
                 "minProperties": 1,
                 "patternProperties": {
-                    "^\w+$": types.NONEMPTY_DICT
+                    r"^\w+$": types.NONEMPTY_DICT
                 }
             },
         }
@@ -173,6 +173,7 @@ class DirectWorkflowSpec(WorkflowSpec):
         self.inbound_tasks_cache = {}
         self.outbound_tasks_cache = {}
 
+    @profiler.trace('direct-wf-spec-validate-semantics', hide_args=True)
     def validate_semantics(self):
         super(DirectWorkflowSpec, self).validate_semantics()
 
@@ -187,6 +188,7 @@ class DirectWorkflowSpec(WorkflowSpec):
         self._check_workflow_integrity()
         self._check_join_tasks()
 
+    @profiler.trace('direct-wf-spec-check-workflow-integrity', hide_args=True)
     def _check_workflow_integrity(self):
         for t_s in self.get_tasks():
             out_task_names = self.find_outbound_task_names(t_s.get_name())
@@ -285,6 +287,9 @@ class DirectWorkflowSpec(WorkflowSpec):
         for tup in self.get_on_complete_clause(task_name):
             t_names.add(tup[0])
 
+        for tup in self.get_on_skip_clause(task_name):
+            t_names.add(tup[0])
+
         return t_names
 
     def transition_exists(self, from_task_name, to_task_name):
@@ -295,7 +300,7 @@ class DirectWorkflowSpec(WorkflowSpec):
     def get_on_error_clause(self, t_name):
         result = []
 
-        on_clause = self.get_tasks()[t_name].get_on_error()
+        on_clause = self.get_task(t_name).get_on_error()
 
         if on_clause:
             result = on_clause.get_next()
@@ -311,10 +316,29 @@ class DirectWorkflowSpec(WorkflowSpec):
 
         return result
 
+    def get_on_skip_clause(self, t_name):
+        result = []
+
+        on_clause = self.get_tasks()[t_name].get_on_skip()
+
+        if on_clause:
+            result = on_clause.get_next()
+
+        if not result:
+            t_defaults = self.get_task_defaults()
+
+            if t_defaults and t_defaults.get_on_skip():
+                result = self._remove_task_from_clause(
+                    t_defaults.get_on_skip().get_next(),
+                    t_name
+                )
+
+        return result
+
     def get_on_success_clause(self, t_name):
         result = []
 
-        on_clause = self.get_tasks()[t_name].get_on_success()
+        on_clause = self.get_task(t_name).get_on_success()
 
         if on_clause:
             result = on_clause.get_next()
@@ -333,7 +357,7 @@ class DirectWorkflowSpec(WorkflowSpec):
     def get_on_complete_clause(self, t_name):
         result = []
 
-        on_clause = self.get_tasks()[t_name].get_on_complete()
+        on_clause = self.get_task(t_name).get_on_complete()
 
         if on_clause:
             result = on_clause.get_next()
@@ -363,7 +387,7 @@ class ReverseWorkflowSpec(WorkflowSpec):
                 "type": "object",
                 "minProperties": 1,
                 "patternProperties": {
-                    "^\w+$": types.NONEMPTY_DICT
+                    r"^\w+$": types.NONEMPTY_DICT
                 }
             },
         }
